
# Tcl/Tk 8.2 script written by Mike Polis and placed in the public domain
# Interpreter available from http://www.scriptics.com
# Ad Majorem Dei Gloriam

# To Do:
#  - load the data from UNIQUES
#  - rename procs and variables in words.tcl to be in a namespace
#  - tag verse text with verse number
#  - communicate with BW via DDE

proc dictAppend {text} {
    global dict
    $dict configure -state normal
    $dict insert end $text
    $dict configure -state disabled
}

# code for using "Words" data
source words.tcl

# Load Lynn Nelson's latwords.aid into $lexicon
proc loadLexicon_aid {filename} {
    global lexicon_aid statline

    if {[catch {set file [open $filename r]} result]} {
	tk_messageBox -icon error -type ok -title {Load Failed} \
		-message "Unable to load lexicon from '$filename': $result"
	return
    }
    set nterms 0
    while {[gets $file line]>=0} {
	if {[regexp {^([^:]+) : } $line dummy terms]} {
	    # enter terms in lexicon
	    foreach term $terms {
		if {[info exists lexicon_aid($term)]} {
		    set lexicon_aid($term) "$lexicon_aid($term)\n$line"
		} else {
		    set lexicon_aid($term) $line
		}
		incr nterms
	    }
	}
    }
    close $file
    set statline "Read $nterms entries from aid lexicon"
}

# return a list of alternating text and tags
proc lookup {word} {
    global lexicon_aid options

    set root [string tolower $word]
    set results {}
    if {[info exists lexicon_aid($root)]} {
	lappend results "$lexicon_aid($root)\n" definition
    }
    set hits [lookup_Words $word]
    if {![llength $hits] && [string match {[A-Z]*} $word]} {
	# try lower case
	set hits [lookup_Words [string tolower $word]]
    }
    # show what was tried if no hit was found
    if {$options(show_misses) && ![llength $hits]} {
	global debug_morphology

	set old_debug_morphology $debug_morphology
	set debug_morphology 1
	set hits [lookup_Words $word]
	if {![llength $hits] && [string match {[A-Z]*} $word]} {
	    # try lower case
	    set hits [lookup_Words [string tolower $word]]
	}
	set debug_morphology $old_debug_morphology
    }
    eval [concat {lappend results} $hits]
    if {![llength $results]} {
	return [list "'$word' not found" definition]
    } else {
	return $results
    }
}

# Load a file into the main text widget
proc load {filename} {
    global text

    if {![string length $filename]} {
	# load was cancelled
	return
    }
    if {[catch {set file [open $filename r]} result]} {
	tk_messageBox -icon error -type ok -title {Load Failed} \
		-message "Unable to load '$filename': $result"
	return
    }
    #$text configure -state normal
    $text delete 1.0 end
    $text insert end [read $file]
    #$text configure -state disabled
    close $file
}
# file types the preceding function knows how to load. The first one will be
# the default in the Open dialog.
set loadtypes {
    {{Text files} {.txt} }
    {{All files} * }
}

# Save the main text widget to a file
proc save {filename} {
    global text

    if {![string length $filename]} {
	# Save was cancelled
	return
    }
    if {[catch {set file [open $filename w]} result]} {
	tk_messageBox -icon error -type ok -title {Load Failed} \
		-message "Unable to save to '$filename': $result"
	return
    }
    # Save all of the text except for the newline which the text widget keeps
    # at the end.
    puts -nonewline $file [$text get 1.0 end-1c]
    close $file
}
# file types the preceding function knows how to save. The first one will be
# the default in the Save dialog.
set savetypes {
    {{Text files} {.txt} }
}

# callback for mouse motion over a word
proc autogloss {source dest x y state} {
    set word [$source get "@$x,$y wordstart" "@$x,$y wordend"]
    # low bit of $state is set if Shift is down
    # we also check to make sure we're over a word and not an empty string
    if {!($state&1) && [regexp {[A-Za-z]} $word]} {
	$dest configure -state normal
	$dest delete 1.0 end
	eval [concat {$dest insert end} [lookup $word]]
	$dest configure -state disabled
    }
}


# callback for change in text widget area shown
# This automatically hides and shows the scroll bar.
proc autobar {widget start end} {
    if {![info exists autobar($widget)]} {
	pack $widget -side right -expand true -fill y
	set autobar($widget) 1
    }
    if {$start==0 && $end==1} {
	# all visible--no scrollbar needed
	if {$autobar($widget)} {
	    set autobar($widget) [expr !$autobar($widget)]
	    pack forget $widget
	}
    } else {
	# only part is visible--scrollbar needed
	if {!$autobar($widget)} {
	    set autobar($widget) [expr !$autobar($widget)]
	    pack $widget -side right -expand true -fill y
	}
	# pass bounds to scrollbar
	$widget set $start $end
    }
}


# create the windows
proc buildgui {} {
    global text dict

    # menu bar
    pack [frame .menubar] -fill x -side top
    pack [menubutton .menubar.file -text File -menu .menubar.file.m] -side left
    set m [menu .menubar.file.m -tearoff false]
    $m add command -label Open... -command \
	    {load [tk_getOpenFile -filetypes $loadtypes]}
    $m add command -label Save... -command \
	    {save [tk_getSaveFile -filetypes $savetypes]}
    $m add command -label Quit -command exit

    pack [menubutton .menubar.edit -text Edit -menu .menubar.edit.m] -side left
    set m [menu .menubar.edit.m -tearoff false]
    $m add command -label Cut -command {event generate $text <<Cut>>}
    $m add command -label Copy -command {event generate $text <<Copy>>}
    $m add command -label Paste -command {event generate $text <<Paste>>}

    # status line (not used)
    #label .statline -textvariable statline -anchor w
    #pack .statline -fill x -side bottom

    # gloss text
    pack [frame .dict] -fill x -side bottom
    scrollbar .dict.s -command {.dict.t yview}
    pack .dict.s -side right -fill y
    set dict [text .dict.t -tabs 3c -wrap word -state disabled -height 9 \
	    -yscrollcommand {autobar .dict.s}]
    $dict tag configure term -font {-family system -weight bold}
    pack $dict -expand true -side left -fill both

    # main text
    pack [frame .text] -expand true -fill both -side top
    scrollbar .text.s -command {.text.t yview}
    pack .text.s -side right -fill y
    set text [text .text.t -wrap word -yscrollcommand {autobar .text.s}]
    pack $text -expand true -fill both -side left

    bind $text <Motion> {autogloss $text $dict %x %y %s}
}

# If this is set to a non-zero number, then a list of failed matches will be
# produced if a word is not found.
set options(show_misses) 0

# do it
wm withdraw .
buildgui
#loadLexicon_aid latwords.aid.txt
loadLexicon_Words
wm deiconify .
