# Code for working with the "Words" Latin lexicon and parser

set debug_morphology 0

# All of the data is stored in one Tcl array (what users of other
# languages would call a hash table)
# 1) map from stems to lexicon entries
#                                  lexicon_Words(stem,$stem) -> list of ids
# 2) lexicon entries               lexicon_Words($id) -> line from file
# 3) map from endings to inflection entries
#                                  lexicon_Words(ending,$ending) -> list of ids
# 4) inflection entries
#                                  lexicon_Words($id,$field) -> data

# Parse fixed width fields from a line.  linename is the name of the
# variable containing the line.  This variable is changed in the
# caller's frame.  id is the id for the lexicon entry, and format is a
# list of pairs of field widths and names.
proc parseFixed {arrayname linename format} {
    upvar $arrayname lexinfo $linename line

    foreach fieldDef $format {
	set len [lindex $fieldDef 0]
	set lexinfo([lindex $fieldDef 1]) \
		[string trim [string range $line 0 [expr $len-1]]]
	set line [string replace $line 0 [expr $len-1]]
    }
}


# Handle a line from dictline.gen
proc loadLexicon_Words_dictline {line id} {
    global lexicon_Words

    set stem1 [string trim [string range $line  0 18]]
    set stem2 [string trim [string range $line 19 37]]
    set stem3 [string trim [string range $line 38 56]]
    set stem4 [string trim [string range $line 57 75]]

    # stick the whole string in the hash table as one chunk
    set lexicon_Words($id) $line

    # index stems
    foreach stem [list $stem1 $stem2 $stem3 $stem4] {
	# only index valid stem strings
	if {[string length $stem] && [string compare $stem zzz]!=0} {
	    if {[string equal stem {.}]} {
		# special case for TO_BE
		set stem {}
	    }
	    if {![info exists lexicon_Words(stem,$stem)]} {
		# there is no list for the stem, so create one
		set lexicon_Words(stem,$stem) $id
	    } elseif {[lsearch $lexicon_Words(stem,$stem) $id]<0} {
		# there's a list and this entry isn't indexed
		# (we need to check because of duplicate stems)
		lappend lexicon_Words(stem,$stem) $id
	    }
	}
    }

    return 1
}


# parse a line from INFLECTS
proc loadLexicon_Words_inflects {line id} {
    global lexicon_Words

    # Remove comments (which start with "--")
    if {[set i [string first -- $line]]>=0} {
	set line [string replace $line $i end]
    }

    # Now we treat it as a Tcl list.  The only problem with this is that
    # there's no stem in the list if the stem length is 0.  We'll handle
    # that as a special case.

    # throw away empty lines
    if {![llength $line]} {
	return 0
    }

    set part_of_speech [lindex $line 0]

    # The next elements depend on the part of speech
    # For several parts of speech, the identification of the kind field
    # is speculative because it's always X in INFLECTS.  Many fields are
    # speculative for SUPINE.
    switch -exact -- $part_of_speech {
	ADV {
	    set inflection_fields {kind}
	}
	PREP {
	    set inflection_fields {case}
	}
	CONJ -
	INTERJ {
	    set inflection_fields {}
	}
	ADJ -
	N {
	    set inflection_fields {
		declination variant case number gender kind
	    }
	}
	NUM {
	    set inflection_fields {
		declination variant case number gender kind unknown
	    }
	}
	PRON {
	    set inflection_fields {
		declination variant case number gender kind
	    }
	}
	V {
	    set inflection_fields {
		declination variant tense voice mood person number kind
	    }
	}
	VPAR {
	    set inflection_fields {
		declination variant case number gender tense voice mood kind
	    }
	}
	SUPINE {
	    set inflection_fields {
		declination variant case number gender kind
	    }
	}
	default {
	    error "Part of speech '$part_of_speech' unimplemented"
	}
    }

    set fields1 [concat {part_of_speech} $inflection_fields {stem endinglen}]
    for {set i 0} {$i < [llength $fields1]} {incr i} {
	if {$i >= [llength $line]} {
	    error "line '$line' does not have enough fields\n"
	}
	set [lindex $fields1 $i] [lindex $line $i]
    }

    # Remove the fields1 part of the list along with the ending, if any
    if {$endinglen > 0} {
	set ending [lindex $line [llength $fields1]]
	set line [lreplace $line 0 [llength $fields1]]
    } else {
	set ending {}
	set line [lreplace $line 0 [expr [llength $fields1]-1]]
    }

    set fields2 {source frequency}
    for {set i 0} {$i < [llength $fields2]} {incr i} {
	if {$i >= [llength $line]} {
	    error "line '$line' does not have enough fields\n"
	}
	set [lindex $fields2 $i] [lindex $line $i]
    }

    # The line was successfully parsed.  Now we add it to the array.

    # index from ending to id
    lappend lexicon_Words(ending,$ending) $id

    # enter fields for id in array (ignore those in skiplist)
    set skiplist {unknown endinglen}
    foreach field [concat $fields1 $fields2] {
	if {[lsearch $skiplist $field]<0} {
	    # for adverbs, X means no degree, not unknown degree.
	    if {[string equal $field kind] && [string equal $kind X] && \
		    [string equal $part_of_speech ADV]} {
		set lexicon_Words($id,kind) NO_DEG
	    } else {
		set lexicon_Words($id,$field) [set $field]
	    }
	}
    }

    # create decoding tables
    lookup_Words_defineTables

    # entry was read successfully
    return 1
}


# load all components
proc loadLexicon_Words {} {
    global lexicon_Words_message
    # list of files and line handler callbacks
    set parsers {
	{INFLECTS loadLexicon_Words_inflects}
	{DICTLINE.GEN loadLexicon_Words_dictline}
    }
    set id 0

    set progress [toplevel .lexicon_Words_progress]
    pack [label $progress.l -textvariable lexicon_Words_message] -side top
    wm title $progress {Loading Words Data}

    foreach parser $parsers {
	global dict

	set filename [lindex $parser 0]
	set lineHandler [lindex $parser 1]

	if {[catch {set file [open $filename r]} result]} {
	    tk_messageBox -icon error -type ok -title {Load Failed} \
		    -message "Unable to open '$filename' for Words lexicon: $result"
	    exit 1
	}
	set lexicon_Words_message "Loading $filename..."
	update
	set nentries 0
	while {[gets $file line]>=0} {
	    if [catch {$lineHandler $line $id} result] {
		global errorInfo
		tk_messageBox -icon error -type ok -title {Load Failed} \
			-message "Error parsing data file $filename for Words lexicon:\n$errorInfo"
		close $file
		exit 1
	    }
	    if {$result} {
		incr nentries
		incr id
		if {$nentries%100==0} {
		    set lexicon_Words_message \
			    "Loaded $nentries entries from $filename"
		    update
		}
	    }
	}
	close $file
    }
    destroy $progress
}


proc lookup_Words_parseDefinition {id arrayname} {
    global lexicon_Words
    upvar $arrayname wordinfo
    set line $lexicon_Words($id)

    parseFixed wordinfo line {
	{19 stem1} {19 stem2} {19 stem3} {19 stem4} {7 part_of_speech}
    }

    # the format of the declination record depends on the part of speech
    set declination_record [string trim [string range $line 0 16]]
    set line [string replace $line 0 16]
    switch $wordinfo(part_of_speech) {
	N {
	    set format {
		{2 declination} {2 variant} {2 gender} {11 kind}
	    }
	}

	ADJ -
	PACK -
	PRON -
	V {
	    set format {
		{2 declination} {2 variant} {13 kind}
	    }
	}

	NUM {
	    set format {
		{2 declination} {2 variant} {7 kind} {6 value}
	    }
	}

	ADV {
	    set format {{17 kind}}
	}

	PREP {
	    set format {{17 case}}
	}

	CONJ -
	INTERJ {
	    set format {}
	}

	default {
	    set pos $wordinfo(part_of_speech)
	    if {![info exists lexicon_Words(complained,$pos)]} {
		global dict

		set lexicon_Words(complained,$pos) 1
		dictAppend "need format for $pos declination record (stem1 = $wordinfo(stem1))\n"
	    }
	    set format {}
	}
    }
    parseFixed wordinfo declination_record $format

    parseFixed wordinfo line {
	{2 age} {2 field} {2 geo} {2 frequency} {2 source}
	{80 definition}
    }

    # if we failed to get a definition, the read failed
    if {[string length $wordinfo(definition)]==0} {
	set info {}
	foreach var [array names wordinfo] {
	    set info [append $info "$var=='$wordinfo($var)'\n"]
	}
	tk_messageBox -icon error -type ok -title {Lexicon format error} \
		-message "parse error for '$lexicon_Words($id)'\n$info"
    }

    # for adverbs, X means no degree, not unknown degree.
    if {[string equal $wordinfo(part_of_speech) ADV] && \
	    [info exists $wordinfo(kind)] && \
	    [string equal $wordinfo(kind) X] } {
	set wordinfo(kind) NO_DEG
    }
}


proc lookup_Words_sort_definition {a b} {
    lookup_Words_parseDefinition [lindex $a 2] wordinfoa
    set afreq $wordinfoa(frequency)
    lookup_Words_parseDefinition [lindex $b 2] wordinfob
    set bfreq $wordinfob(frequency)
    # map unknown values to C (the word would be in a 10,000 word dictionary)
    if {[string equal $afreq X]} {
	set afreq C
    }
    if {[string equal $bfreq X]} {
	set bfreq C
    }
    if {$afreq < $bfreq} {
	return -1
    } elseif {$afreq > $bfreq} {
	return 1
    } else {
	return 0
    }
}


proc decode {field data} {
    global lexicon_Words

    if {[info exists lexicon_Words($field,$data)]} {
	return $lexicon_Words($field,$data)
    } else {
	return $data
    }
}


# parse and look up a word in Words and return a list of hits.  Each hit is
# a list ($results):
#  0) the stem and ending separated by a "."
#  1) a list of descriptions of morphologies which would lead to that form
#  2) id of the dictionary entry information for the word
proc lookup_Words_list {word} {
    global lexicon_Words
    global debug_morphology

    set results {}
    set wordlen [string length $word]

    # try every possible split into stem and ending
    for {set stemlen $wordlen} {$stemlen>=0} {incr stemlen -1} {
	if {$stemlen==0} {
	    # special case needed for 2nd TO_BE stem (e.g. erat)
	    set stem .
	} else {
	    set stem [string range $word 0 [expr $stemlen-1]]
	}
	set ending [string range $word $stemlen end]
	if {[info exists lexicon_Words(stem,$stem)] && \
		[info exists lexicon_Words(ending,$ending)]} {
	    set declinations $lexicon_Words(ending,$ending)
	    foreach definition $lexicon_Words(stem,$stem) {
		lookup_Words_parseDefinition $definition wordinfo
		set decl_hits {}
		foreach d $declinations {
		    # determine whether ending can be used with stem
		    if {[string equal $wordinfo(part_of_speech) \
			    $lexicon_Words($d,part_of_speech)]} {
			set pos $wordinfo(part_of_speech)
			set morphology [decode part_of_speech $pos]
		    } elseif {[string equal $wordinfo(part_of_speech) V] && \
			    [string equal \
			    $lexicon_Words($d,part_of_speech) VPAR]} {
			set pos VPAR
			set morphology verb
		    } elseif {$debug_morphology} {
			set pos X
			set morphology "$wordinfo(part_of_speech)!=$lexicon_Words($d,part_of_speech)"
		    } else {
			continue
		    }

		    if {![string equal $stem \
			    $wordinfo(stem$lexicon_Words($d,stem))]} {
			if {$debug_morphology} {
			    append morphology \
				    " $stem!=stem$lexicon_Words($d,stem)"
			} else {
			    continue
			}
		    }

		    if {[info exists lexicon_Words($d,declination)] && \
			    [info exists wordinfo(declination)] && \
			    !($lexicon_Words($d,declination) == \
			    $wordinfo(declination) || \
			    $lexicon_Words($d,declination)==0)} {
			if {$debug_morphology} {
			    append morphology " $lexicon_Words($d,declination)!=$wordinfo(declination)"
			} else {
			    continue
			}
		    }

		    if {[info exists lexicon_Words($d,variant)] && \
			    [info exists wordinfo(variant)] && \
			    !($lexicon_Words($d,variant) == \
			    $wordinfo(variant) || \
			    $lexicon_Words($d,variant)==0)} {
			if {$debug_morphology} {
			    append morphology " $lexicon_Words($d,variant)!=$wordinfo(variant)"
			} else {
			    continue
			}
		    }

		    if {[info exists lexicon_Words($d,kind)] && \
			    [info exists wordinfo(kind)]} {
			if {[string equal $lexicon_Words($d,kind) \
				$wordinfo(kind)] || \
				[string equal $lexicon_Words($d,kind) X]} {
			    if {![string equal $wordinfo(kind) X]} {
				append morphology \
					" [decode kind $wordinfo(kind)]"
			    }
			} elseif {[string equal $wordinfo(kind) X]} {
			    append morphology \
				    " [decode kind $lexicon_Words($d,kind)]"
			} elseif {$debug_morphology} {
			    append morphology \
				    " $lexicon_Words($d,kind)!=$wordinfo(kind)"
			} else {
			    continue
			}
		    }

		    if {[info exists wordinfo(value)] && $wordinfo(value)!=0} {
			# value of a number
			append morphology " $wordinfo(value)"
		    }

		    foreach field {mood voice tense case} {
			if {[info exists lexicon_Words($d,$field)]} {
			    append morphology " [decode $field $lexicon_Words($d,$field)]"
			}
		    }

		    if {[info exists lexicon_Words($d,gender)] && \
			    [info exists wordinfo(gender)]} {
			set gender_a $lexicon_Words($d,gender)
			set gender_b $wordinfo(gender)
			if {[string equal $gender_a $gender_b] || \
				[string equal $gender_b X] || \
				(![string equal $gender_a N] && \
				[string equal $gender_b C])} {
			    append morphology " [decode gender $gender_a]"
			} elseif {[string equal $gender_a X] || \
				(![string equal $gender_b N] && \
				[string equal $gender_a C])} {
			    append morphology " [decode gender $gender_b]"
			} elseif {$debug_morphology} {
			    append morphology " $gender_a!=$gender_b"
			} else {
			    continue
			}
		    } elseif {[string equal $pos VPAR] && \
			    [info exists lexicon_Words($d,gender)]} {
			append morphology \
				" [decode gender $lexicon_Words($d,gender)]"
		    }

		    if {[info exists lexicon_Words($d,person)]} {
			append morphology \
				" [decode person $lexicon_Words($d,person)]"
		    }

		    if {[info exists lexicon_Words($d,number)]} {
			append morphology \
				" [decode number $lexicon_Words($d,number)]"
		    }

		    if {[info exists lexicon_Words(endingfreq,$lexicon_Words($d,frequency))]} {
			append morphology " ($lexicon_Words(endingfreq,$lexicon_Words($d,frequency)))"
		    }

		    lappend decl_hits $morphology
		}
		# if we found at least one declination, add to results
		if {[llength $decl_hits]} {
		    if {$stemlen==0} {
			set stem {}
		    }
		    lappend results \
			    [list $stem.$ending $decl_hits $definition]
		}
	    }
	}
    }

    return [lsort -command lookup_Words_sort_definition $results]
}


# Take lookup_Words_list's results and format them neatly, returning a list
# of alternating text and text widget tags.
proc lookup_Words {word} {
    global lexicon_Words

    set results {}
    foreach hit [lookup_Words_list $word] {
	lappend results [lindex $hit 0] term
	foreach form [lindex $hit 1] {
	    lappend results "\t$form\n" form
	}
	lookup_Words_parseDefinition [lindex $hit 2] wordinfo
	set extra {}
	if {![string equal $wordinfo(frequency) X] && \
		[string compare $wordinfo(frequency) E]>=0} {
	    lappend extra [decode frequency $wordinfo(frequency)]
	}
	foreach field {field age geo source} {
	    if {![string equal $wordinfo($field) X]} {
		lappend extra [decode $field $wordinfo($field)]
	    }
	}
	lappend results $wordinfo(definition) definition
	if {[llength $extra]} {
	    lappend results " ([join $extra {; }])" definition_extra
	}
	lappend results "\n" definition
    }

    return $results
}

# function to set up decoding tables for root and ending fields
proc lookup_Words_defineTables {} {
    global lexicon_Words

    set tables {
	{source A {Allen + Greenough, New Latin Grammar, 1888}}
	{source B {J.T.Bretzke, Consecrated Phrases: Theolog Dict}}
	{source C {Cassell's Latin Dictionary 1968}}
	{source D {J.N.Adams, Latin Sexual Vocabulary, 1982}}
	{source E {L.F.Stelten, Dictionary of Eccles. Latin, 1995}}
	{source F {Roy J. Deferrari, Dictionary of St. Thomas Aquinas, 1960}}
	{source G {Gildersleeve + Lodge, Latin Grammar 1895}}
	{source H {Harrington/Pucci/Elliott, Medieval Latin 2nd Ed 1997}}
	{source I {Leverett, F.P., Lexicon of the Latin Language, Boston 1845}}
	{source K {W. M. Lindsay, Short Historical Latin Grammar, 1895}}
	{source L {Lewis, C.S., Elementary Latin Dictionary 1891}}
	{source M {Latham, Revised Medieval Word List, 1980}}
	{source N {Lynn Nelson, Wordlist}}
	{source O {Oxford Latin Dictionary, 1982}}
	{source Q {Other, unspecified dictionaries}}
	{source S {Lewis and Short, A Latin Dictionary, 1879}}
	{source T {Found in a translation -- no dictionary reference}}
	{source U {Du Cange}}
	{source V {Vademecum in opus Saxonis - Franz Blatt}}
	{source W {William Whitaker's personal guess}}
	{source Y {Niermeyer, Mediae Latinitatis Lexicon Minus}}
	{source Z {Sent by Words user -- no dictionary reference}}
	{age A {Archaic Latin}}
	{age B {Early Latin}}
	{age C {Classical Latin}}
	{age D {Late Latin}}
	{age E {Later Latin}}
	{age F {Medieval Latin}}
	{age G {Modern Latin}}
	{age H {Neo-Latin}}
	{field A {Agriculture, Flora, Fauna, Land, Equipment, Rural}}
	{field B {Biological, Medical, Body Parts}}
	{field D {Drama, Music, Theater, Art, Painting, Sculpture}}
	{field E {Ecclesiastic, Biblical, Religious}}
	{field G {Grammar, Rhetoric, Logic, Literature, Schools}}
	{field L {Legal, Government, Tax, Financial, Political, Titles}}
	{field P {Poetic}}
	{field S {Science, Philosophy, Mathematics, Units/Measures}}
	{field T {Technical, Architecture, Topography, Surveying}}
	{field W {War, Military, Naval, Armor}}
	{field Y {Mythology}}
	{geo A {Africa}}
	{geo B {Britain}}
	{geo C {China}}
	{geo D {Scandinavia}}
	{geo E {Egypt}}
	{geo F {France, Gaul}}
	{geo G {Germany}}
	{geo H {Greece}}
	{geo I {Italy, Rome}}
	{geo J {India}}
	{geo K {Balkans}}
	{geo N {Netherlands}}
	{geo P {Persia}}
	{geo Q {Near East}}
	{geo R {Russia}}
	{geo S {Spain, Iberia}}
	{geo U {Eastern Europe}}
	{geo Y {Mythology}}
	{frequency A {very frequent}}
	{frequency B {frequent}}
	{frequency C {common}}
	{frequency D {lesser}}
	{frequency E {uncommon}}
	{frequency F {very rare}}
	{frequency I {only in inscriptions}}
	{frequency M {only in graffiti}}
	{frequency N {only in Pliny}}
	{endingfreq B {variant form}}
	{endingfreq C {uncommon form}}
	{endingfreq D {infrequent form}}
	{endingfreq E {rare form}}
	{endingfreq F {very rare form}}
	{endingfreq I {only in inscriptions}}
	{endingfreq M {only in graffiti}}
	{part_of_speech X {all, none, or unknown}}
	{part_of_speech N {noun}}
	{part_of_speech PRON {pronoun}}
	{part_of_speech PACK {PACKON -- artificial for code}}
	{part_of_speech ADJ {adjective}}
	{part_of_speech NUM {numeral}}
	{part_of_speech ADV {adverb}}
	{part_of_speech V {verb}}
	{part_of_speech VPAR {verb participle}}
	{part_of_speech SUPINE {supine}}
	{part_of_speech PREP {preposition}}
	{part_of_speech CONJ {conjunction}}
	{part_of_speech INTERJ {interjection}}
	{part_of_speech TACKON {TACKON -- artificial for code}}
	{part_of_speech PREFIX {PREFIX --  here artificial for code}}
	{part_of_speech SUFFIX {SUFFIX --  here artificial for code}}
	{gender X {gender unknown}}
	{gender M {masculine}}
	{gender F {feminine}}
	{gender N {neuter}}
	{gender C {common}}
	{case X {case unknown}}
	{case NOM {nominative}}
	{case VOC {vocative}}
	{case GEN {genitive}}
	{case LOC {locative}}
	{case DAT {dative}}
	{case ABL {ablative}}
	{case ACC {accusative}}
	{person 0 {person unknown}}
	{person 1 {1st person}}
	{person 2 {2nd person}}
	{person 3 {3rd person}}
	{number X {number unknown}}
	{number S {singular}}
	{number P {plural}}
	{tense X {tense: all, none, or unknown}}
	{tense PRES {present}}
	{tense IMPF {imperfect}}
	{tense FUT {future}}
	{tense PERF {perfect}}
	{tense PLUP {pluperfect}}
	{tense FUTP {future perfect}}
	{voice X {voice: all, none, or unknown}}
	{voice ACTIVE {active}}
	{voice PASSIVE {passive}}
	{mood X {mood: all, none, or unknown}}
	{mood IND {indicative}}
	{mood SUB {subjunctive}}
	{mood IMP {imperative}}
	{mood INF {infinitive}}
	{mood PPL {participle}}
	{kind A {(abstract idea)}}
	{kind ABL {(taking the ablative)}}
	{kind ADJECT {adjectival}}
	{kind ADVERB {numeral adverb}}
	{kind CARD {cardinal}}
	{kind COMP {comparative}}
	{kind DAT {(taking the dative)}}
	{kind DEMONS {demonstrative}}
	{kind DEP {(deponent)}}
	{kind DIST {distributive}}
	{kind GEN {(taking the genitive)}}
	{kind IMPERS {(impersonal)}}
	{kind INDEF {indefinite}}
	{kind INTERR {interrogative}}
	{kind INTRANS {(intransitive)}}
	{kind L {(place name)}}
	{kind M {(plural or multiple "only")}}
	{kind N {(proper name)}}
	{kind NO_DEG {no degree}}
	{kind ORD {ordinal}}
	{kind P {(person)}}
	{kind PERFDEF {(perfect definite)}}
	{kind PERS {personal}}
	{kind REFLEX {reflexive}}
	{kind POS {positive}}
	{kind REL {relative}}
	{kind S {(singular "only")}}
	{kind SEMIDEP {(semideponent)}}
	{kind SUPER {superlative}}
	{kind T {(thing)}}
	{kind TO_BE {(to be)}}
	{kind TO_BEING {(compound of to be)}}
	{kind TRANS {(transitive)}}
	{kind W {(place)}}
    }

    foreach ent $tables {
	set lexicon_Words([lindex $ent 0],[lindex $ent 1]) [lindex $ent 2]
    }
}
